
package shopping.cart;

import akka.Done;
import akka.projection.eventsourced.EventEnvelope;
import akka.projection.javadsl.Handler;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.CompletionStage;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

public final class ItemPopularityProjectionHandler
    extends Handler<EventEnvelope<ShoppingCart.Event>> { 
  private final Logger logger = LoggerFactory.getLogger(getClass());
  private final String tag;
  private final ItemPopularityRepository repo;

  public ItemPopularityProjectionHandler(String tag, ItemPopularityRepository repo) {
    this.tag = tag;
    this.repo = repo;
  }

  @Override
  public CompletionStage<Done> process(EventEnvelope<ShoppingCart.Event> envelope) { 
    ShoppingCart.Event event = envelope.event();

    switch (event) {
      case ShoppingCart.ItemAdded(var __, String itemId, int qtd) -> { 
        CompletionStage<Done> result = this.repo.update(itemId, qtd);
        result.thenAccept(done -> logItemCount(itemId));
        return result;
        
      }
      case ShoppingCart.ItemQuantityAdjusted(var __, String itemId, int oldQtd, int newQtd) -> {
        CompletionStage<Done> result = this.repo.update(itemId, newQtd - oldQtd);
        result.thenAccept(done -> logItemCount(itemId));
        return result;
      }
      case ShoppingCart.ItemRemoved(var __, String itemId, int oldQtd) -> {
        CompletionStage<Done> result = this.repo.update(itemId, -oldQtd);
        result.thenAccept(done -> logItemCount(itemId));
        return result;
      }
        
      case ShoppingCart.CheckedOut ignored -> {
        return CompletableFuture.completedFuture(Done.getInstance());
      }
      case null -> throw new IllegalArgumentException("Unknown event type: " + event.getClass());
    }
  }

  private void logItemCount(String itemId) {
    repo.getItem(itemId)
        .thenAccept(
            optCount ->
                logger.info(
                    "ItemPopularityProjectionHandler({}) item popularity for '{}': [{}]",
                    this.tag,
                    itemId,
                    optCount.orElse(0L)));
  }
}

