
package shopping.cart;

import akka.Done;
import akka.NotUsed;
import akka.actor.typed.ActorSystem;
import akka.kafka.ProducerMessage;
import akka.kafka.ProducerSettings;
import akka.kafka.javadsl.Producer;
import akka.projection.ProjectionContext;
import akka.projection.eventsourced.EventEnvelope;
import akka.stream.javadsl.FlowWithContext;
import com.google.protobuf.Any;
import org.apache.kafka.clients.producer.ProducerRecord;
import org.apache.kafka.common.serialization.ByteArraySerializer;
import org.apache.kafka.common.serialization.StringSerializer;

public final class PublishEventsProjectionFlow {

  public static FlowWithContext<
          EventEnvelope<ShoppingCart.Event>, ProjectionContext, Done, ProjectionContext, NotUsed>
      createProducerFlow(String topic, ActorSystem<?> system) {
    var producerSettings =
        ProducerSettings.create(system, new StringSerializer(), new ByteArraySerializer());

    
    return FlowWithContext.<EventEnvelope<ShoppingCart.Event>, ProjectionContext>create()
        .map(
            envelope -> {
              ShoppingCart.Event event = envelope.event();

              // using the cartId as the key and `DefaultPartitioner` will select partition based on
              // the key so that events for same cart always ends up in same partition
              String key = event.cartId();
              var producerRecord = new ProducerRecord<>(topic, key, serialize(event)); 
              system.log().info("Publishing event [{}] to topic {}", envelope.event(), topic);
              return ProducerMessage.single(producerRecord);
            })
        .via(Producer.flowWithContext(producerSettings))
        .map(__ -> Done.getInstance());
  }

  private static byte[] serialize(ShoppingCart.Event event) {
    var pbEvent =
        switch (event) {
          case ShoppingCart.ItemAdded(String cartId, String itemId, int qtd) ->
              shopping.cart.proto.ItemAdded.newBuilder()
                  .setCartId(cartId)
                  .setItemId(itemId)
                  .setQuantity(qtd)
                  .build();
          case ShoppingCart.CheckedOut(String cartId, var __) ->
              shopping.cart.proto.CheckedOut.newBuilder().setCartId(cartId).build();
            
          case ShoppingCart.ItemQuantityAdjusted(var cartId, var itemId, int __, int newQtd) ->
              shopping.cart.proto.ItemQuantityAdjusted.newBuilder()
                  .setCartId(cartId)
                  .setItemId(itemId)
                  .setQuantity(newQtd)
                  .build();
          case ShoppingCart.ItemRemoved(String cartId, String itemId, int __) ->
              shopping.cart.proto.ItemRemoved.newBuilder()
                  .setCartId(cartId)
                  .setItemId(itemId)
                  .build();
            
          case null ->
              throw new IllegalArgumentException("Unknown event type: " + event.getClass());
        };

    // pack in Any so that type information is included for deserialization
    return Any.pack(pbEvent).toByteArray(); 
  }
}

